import logging

from celery.task.control import revoke
from cliff.command import Command


class Run(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument('task', nargs=1, type=str, help='Task to revoke')
        return parser

    def take_action(self, parsed_args):
        task = parsed_args.role[0]

        revoke(task, terminate=True)



import logging
import subprocess

from cliff.command import Command


class Run(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument('--type', default="ssh", help='Type of the console')
        parser.add_argument('target', nargs=1, type=str, help='Hostname or address of the console to connect')
        return parser

    def take_action(self, parsed_args):
        type_console = parsed_args.type
        target = parsed_args.target[0]

        if type_console == "ansible":
            subprocess.call(f"/run-ansible-console.sh {target}", shell=True)
        elif type_console == "ssh":
            # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
            subprocess.call(f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator -o StrictHostKeyChecking=no dragon@{target}", shell=True)
        elif type_console == "container":
            target_containername = target.split("/")[1]
            target_hostname = target.split("/")[0]
            target_command = "bash"

            ssh_command = f"docker exec -it {target_containername} {target_command}"
            ssh_options = "-o RequestTTY=force -o StrictHostKeyChecking=no"

            # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
            subprocess.call(f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{target_hostname} {ssh_command}", shell=True)



import logging

from pottery import Redlock

from osism import utils
from osism.plugins import routeros, routeros_testing


def for_device(name, parameters={}):

    device = utils.nb.dcim.devices.get(name=name)

    if "device_type" not in device.custom_fields or device.custom_fields["device_type"] != "switch":
        return

    if "Managed by OSISM" not in [str(x) for x in device.tags]:
        return

    if "deployment_enabled" in device.custom_fields and not bool(device.custom_fields["deployment_enabled"]):
        return

    if "deployment_type" not in device.custom_fields:
        return

    # Allow only one change per time
    lock = Redlock(key=f"lock_check_{name}", masters={utils.redis}, auto_release_time=120)
    lock.acquire()

    logging.info(f"Check configuration for device {device.name} with plugin {device.custom_fields['deployment_type']}")

    if device.custom_fields["deployment_type"] == "routeros":
        last_configuration = routeros.get_configuration(device)
    elif device.custom_fields["deployment_type"] == "routeros_testing":
        last_configuration = routeros_testing.get_configuration(device)
    else:
        logging.error(f"Deployment type x for device {device.name} not supported")
        last_configuration = None

    if last_configuration:
        for line in last_configuration.split('\n'):
            logging.info(f"configuration - {device.name}: {line}")

    lock.release()



import logging
import subprocess

from cliff.command import Command


class Run(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument('target', nargs=1, type=str, help='Hostname or address')
        return parser

    def take_action(self, parsed_args):
        target_hostname = parsed_args.target[0]

        ssh_command = "docker ps"
        ssh_options = "-o StrictHostKeyChecking=no"

        # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
        subprocess.call(f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{target_hostname} {ssh_command}", shell=True)



import logging
import os

from netmiko import ConnectHandler
from paramiko import AutoAddPolicy, SSHClient
from routeros_diff.parser import RouterOSConfig
from scp import SCPClient


def get_netmiko_connection(device):
    parameters = get_parameters(device)
    result = ConnectHandler(device_type="mikrotik_routeros", **parameters)
    return result


def get_scp_connection(device):
    parameters = get_parameters(device)

    ssh_parameters = {
        "hostname": parameters["host"],
        "username": parameters["username"],
        "password": parameters["password"]
    }

    ssh = SSHClient()
    ssh.set_missing_host_key_policy(AutoAddPolicy())
    ssh.connect(**ssh_parameters)
    result = SCPClient(ssh.get_transport())
    return result


def get_configuration(device):
    conn = get_netmiko_connection(device)
    conn.send_command(f"/export file={device.name}")

    scp = get_scp_connection(device)
    scp.get(f"/{device.name}.rsc", f"/tmp/{device.name}.rsc")
    scp.close()

    with open(f"/tmp/{device.name}.rsc", 'r') as fp:
        result = fp.read()

    os.remove(f"/tmp/{device.name}.rsc")

    return result


def get_parameters(device):
    # FIXME: use get_context_data() in the future
    config_context = device.local_context_data

    result = {
        'host': config_context['deployment_address'],
        'username': config_context['deployment_user'],
        'password': config_context['deployment_password']
    }

    return result


def deploy(device, current_configuration, last_configuration):
    if not last_configuration:
        last_configuration = get_configuration(device)

    last = RouterOSConfig.parse(last_configuration)
    current = RouterOSConfig.parse(current_configuration)

    diff = str(current.diff(last))

    if diff:
        for line in diff.split('\n'):
            logging.info(f"diff - {device.name}: {line}")


def diff(device, current_configuration, last_configuration):
    if not last_configuration:
        last_configuration = get_configuration(device)

    last = RouterOSConfig.parse(last_configuration)
    current = RouterOSConfig.parse(current_configuration)

    diff = str(current.diff(last))
    for line in diff.split('\n'):
        logging.info(f"diff - {device.name}: {line}")



import logging
import subprocess

from cliff.command import Command


class Run(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument('type', nargs=1, type=str, help='Type of the worker')
        return parser

    def take_action(self, parsed_args):
        queue = parsed_args.type[0]

        if queue in ["openstack", "netbox", "conductor"]:
            tasks = queue
        else:
            tasks = queue[:-8]

        # NOTE: use python interface in the future, works for the moment
        if tasks == "osism":
            p = subprocess.Popen(f"celery -A osism.tasks.ansible worker -n {queue} --loglevel=INFO -Q {queue}", shell=True)
        else:
            p = subprocess.Popen(f"celery -A osism.tasks.{tasks} worker -n {queue} --loglevel=INFO -Q {queue}", shell=True)
        p.wait()



import logging

from celery import Celery
from cliff.command import Command
from tabulate import tabulate

from osism.tasks import Config

# https://stackoverflow.com/questions/4048651/python-function-to-convert-seconds-into-minutes-hours-and-days/4048773

INTERVALS = (
    ('weeks', 604800),  # 60 * 60 * 24 * 7
    ('days', 86400),    # 60 * 60 * 24
    ('hours', 3600),    # 60 * 60
    ('minutes', 60),
    ('seconds', 1),
)


def display_time(seconds, granularity=2):
    result = []

    for name, count in INTERVALS:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{} {}".format(value, name))
    return ', '.join(result[:granularity])


class Run(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument('type', nargs=1, type=str, help='Type of resource from which the status is to be displayed')
        return parser

    def take_action(self, parsed_args):
        type_of_resource = parsed_args.type[0]

        app = Celery('status')
        app.config_from_object(Config)

        if type_of_resource == "workers":
            table = []
            i = app.control.inspect()
            s = i.stats()
            for node in sorted(s.keys()):
                table.append([
                    node,
                    display_time(s[node]["uptime"])
                ])

            print(tabulate(table, headers=["Name", "Uptime"], tablefmt="psql"))
        else:
            logging.error(f"Unknown resource type '{type_of_resource}'")



import os
import logging

import json
from kombu import BrokerConnection
from kombu import Exchange
from kombu import Queue
from kombu.mixins import ConsumerMixin

from osism import utils
from osism.tasks import netbox, openstack

EXCHANGE_NAME = "ironic"
ROUTING_KEY = "ironic_versioned_notifications.info"
QUEUE_NAME = "osism-listener-ironic"
BROKER_URI = os.getenv("BROKER_URI")

logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')


class NotificationsDump(ConsumerMixin):

    def __init__(self, connection):
        self.connection = connection
        return

    def get_consumers(self, consumer, channel):
        exchange = Exchange(EXCHANGE_NAME, type="topic", durable=False)
        queue = Queue(QUEUE_NAME, exchange, routing_key=ROUTING_KEY, durable=False, auto_delete=True, no_ack=True)
        return [consumer(queue, callbacks=[self.on_message])]

    def on_message(self, body, message):
        data = json.loads(body["oslo.message"])
        # logging.info(data)

        event_type = data["event_type"]
        payload = data["payload"]
        object_data = payload["ironic_object.data"]

        name = object_data["name"]

        # References:
        #
        # * https://docs.openstack.org/ironic/latest/admin/notifications.html#available-notifications
        # * https://docs.openstack.org/ironic/latest/user/states.html
        # * https://docs.openstack.org/ironic/latest/_images/states.svg

        if event_type == "baremetal.node.power_set.end":
            logging.info(f"baremetal.node.power_set.end ## {name} ## {object_data['power_state']}")
            netbox.set_state.delay(name, object_data['power_state'], "power")

        elif event_type == "baremetal.node.power_state_corrected.success":
            logging.info(f"baremetal.node.power_state_corrected.success ## {name} ## {object_data['power_state']}")
            netbox.set_state.delay(name, object_data['power_state'], "power")

        elif event_type == "baremetal.node.maintenance_set.end":
            logging.info(f"baremetal.node.maintenance_set.end ## {name} ## {object_data['maintenance']}")
            netbox.set_maintenance.delay(name, object_data['maintenance'])

        elif event_type == "baremetal.node.provision_set.start":
            logging.info(f"baremetal.node.provision_set.start ## {name} ## {object_data['provision_state']}")

            if object_data["event"] == "inspect":
                # system should be in state a
                netbox.connect.delay(name, "a")

            if object_data["provision_state"] == "cleaning":
                # system should be in state b
                netbox.connect.delay(name, "b")

            if object_data["provision_state"] == "available":
                # system should be in state c
                netbox.connect.delay(name, "c")

            if object_data["target_provision_state"] == "active":
                pass

        # A provision status was successfully set, update it in the netbox
        elif event_type == "baremetal.node.provision_set.success":
            logging.info(f"baremetal.node.provision_set.success ## {name} ## {object_data['provision_state']}")
            netbox.set_state.delay(name, object_data['provision_state'], "provision")

            if object_data["provision_state"] == "manageable":

                # system should be in state c
                netbox.connect.delay(name, "c")

        elif event_type == "baremetal.node.provision_set.end":
            logging.info(f"baremetal.node.provision_set.end ## {name} ## {object_data['provision_state']}")
            netbox.set_state.delay(name, object_data['provision_state'], "provision")

            if object_data["previous_provision_state"] == "inspect wait" and \
               object_data["event"] == "done":
                netbox.set_state.delay(name, "introspected", "introspection")
                openstack.baremetal_set_node_provision_state.delay(name, "provide")

            elif object_data["previous_provision_state"] == "wait call-back":
                pass

            elif object_data["previous_provision_state"] == "cleaning" and \
                 object_data["provision_state"] == "available":  # noqa

                # system should be in state c
                netbox.connect.delay(name, "c")

        elif event_type == "baremetal.port.create.end":
            logging.info(f"baremetal.port.create.end ## {object_data['uuid']}")

            mac_address = object_data["address"]
            interface_a = utils.nb.dcim.interfaces.get(mac_address=mac_address)
            device_a = interface_a.device

            task = openstack.baremetal_get_network_interface_name.delay(device_a.name, mac_address)
            task.wait(timeout=None, interval=0.5)
            network_interface_name = task.get()

            netbox.update_network_interface_name.delay(object_data["address"], network_interface_name)

        elif event_type == "baremetal.port.update.end":
            logging.info(f"baremetal.port.update.end ## {object_data['uuid']}")

            mac_address = object_data["address"]
            interface_a = utils.nb.dcim.interfaces.get(mac_address=mac_address)
            device_a = interface_a.device

            task = openstack.baremetal_get_network_interface_name.delay(device_a.name, mac_address)
            task.wait(timeout=None, interval=0.5)
            network_interface_name = task.get()

            netbox.update_network_interface_name.delay(object_data["address"], network_interface_name)

        elif event_type == "baremetal.node.delete.end":
            logging.info(f"baremetal.node.delete.end ## {name}")

            netbox.set_state.delay(name, "unregistered", "ironic")
            netbox.set_state.delay(name, None, "provision")
            netbox.set_state.delay(name, None, "power")
            netbox.set_state.delay(name, None, "introspection")
            netbox.set_state.delay(name, None, "deployment")

            # system should be in state a
            netbox.connect.delay(name, "a")

            # remove internal flavor
            openstack.baremetal_delete_internal_flavor.delay(name)

        else:
            logging.info(f"{event_type} ## {name}")

        logging.info(object_data)


def main():
    with BrokerConnection(BROKER_URI, connect_timeout=30.0) as connection:
        NotificationsDump(connection).run()


if __name__ == "__main__":
    main()



from celery import Celery
from celery.signals import worker_process_init
import keystoneauth1
import openstack
from redis import Redis
import yaml

from osism.tasks import Config

app = Celery('conductor')
app.config_from_object(Config)


configuration = {}
redis = None


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global configuration
    global redis

    redis = Redis(host="redis", port="6379")

    # Parameters come from the environment, OS_*
    try:
        conn = openstack.connect()
    except keystoneauth1.exceptions.auth_plugins.MissingRequiredOptions:
        pass

    with open("/etc/conductor.yml") as fp:
        configuration = yaml.load(fp, Loader=yaml.SafeLoader)

        # Resolve all IDs in the conductor.yml
        if Config.enable_ironic in ["True", "true", "Yes", "yes"]:
            # TODO: use osism.tasks.openstack in the future
            if "driver_info" in configuration["ironic_parameters"]:
                if "deploy_kernel" in configuration["ironic_parameters"]["driver_info"]:
                    result = conn.image.find_image(configuration["ironic_parameters"]["driver_info"]["deploy_kernel"])
                    configuration["ironic_parameters"]["driver_info"]["deploy_kernel"] = result.id

                if "deploy_ramdisk" in configuration["ironic_parameters"]["driver_info"]:
                    result = conn.image.find_image(configuration["ironic_parameters"]["driver_info"]["deploy_ramdisk"])
                    configuration["ironic_parameters"]["driver_info"]["deploy_ramdisk"] = result.id

                if "cleaning_network" in configuration["ironic_parameters"]["driver_info"]:
                    result = conn.network.find_network(configuration["ironic_parameters"]["driver_info"]["cleaning_network"])
                    configuration["ironic_parameters"]["driver_info"]["cleaning_network"] = result.id

                if "provisioning_network" in configuration["ironic_parameters"]["driver_info"]:
                    result = conn.network.find_network(configuration["ironic_parameters"]["driver_info"]["provisioning_network"])
                    configuration["ironic_parameters"]["driver_info"]["provisioning_network"] = result.id


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.conductor.get_ironic_parameters")
def get_ironic_parameters(self):
    if "ironic_parameters" in configuration:
        return configuration["ironic_parameters"]

    return {}



from setuptools import find_packages
from setuptools import setup

PROJECT = 'osism'
exec(open(f'{PROJECT}/version.py').read())

try:
    long_description = open('README.md', 'rt').read()
except IOError:
    long_description = ''

setup(
    name=PROJECT,
    version=__version__,  # noqa

    description='OSISM manager interface',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='OSISM GmbH',
    author_email='info@osism.tech',

    maintainer='OSISM GmbH',
    maintainer_email='info@osism.tech',

    url='https://github.com/osism/python-osism',
    download_url='https://github.com/osism/python-osism/tarball/main',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Environment :: Console',
    ],

    platforms=['Any'],

    scripts=[],

    provides=[],
    install_requires=['cliff'],

    namespace_packages=[],
    packages=find_packages(),
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'osism = osism.main:main'
        ],
        'osism.commands': [
            'apply = osism.commands.apply:Run',
            'bifrost = osism.commands.bifrost:Run',
            'bifrost deploy = osism.commands.bifrost:Deploy',
            'compose = osism.commands.compose:Run',
            'console = osism.commands.console:Run',
            'container = osism.commands.container:Run',
            'netbox = osism.commands.netbox:Run',
            'netbox check = osism.commands.netbox:Check',
            'netbox connect = osism.commands.netbox:Connect',
            'netbox deploy= osism.commands.netbox:Deploy',
            'netbox diff = osism.commands.netbox:Diff',
            'netbox disable = osism.commands.netbox:Disable',
            'netbox generate = osism.commands.netbox:Generate',
            'netbox import = osism.commands.netbox:Import',
            'netbox init = osism.commands.netbox:Init',
            'netbox manage = osism.commands.netbox:Manage',
            'netbox ping = osism.commands.netbox:Ping',
            'netbox sync = osism.commands.netbox:Sync',
            'netbox sync bifrost = osism.commands.netbox:Bifrost',
            'netbox sync ironic = osism.commands.netbox:Ironic',
            'reconciler = osism.commands.reconciler:Run',
            'reconciler sync = osism.commands.reconciler:Sync',
            'revoke = osism.commands.revoke:Run',
            'service = osism.commands.service:Run',
            'status = osism.commands.status:Run',
            'worker = osism.commands.worker:Run'
        ]
    },

    zip_safe=False,
)



from osism import settings
import pynetbox
from redis import Redis

redis = Redis(host="redis", port="6379")
nb = pynetbox.api(
    settings.NETBOX_URL,
    token=settings.NETBOX_TOKEN
)

if settings.IGNORE_SSL_ERRORS:
    import requests
    requests.packages.urllib3.disable_warnings()
    session = requests.Session()
    session.verify = False
    nb.http_session = session


# https://stackoverflow.com/questions/2361426/get-the-first-item-from-an-iterable-that-matches-a-condition
def first(iterable, condition=lambda x: True):
    """
    Returns the first item in the `iterable` that
    satisfies the `condition`.

    If the condition is not given, returns the first item of
    the iterable.

    Raises `StopIteration` if no item satysfing the condition is found.

    >>> first( (1,2,3), condition=lambda x: x % 2 == 0)
    2
    >>> first(range(3, 100))
    3
    >>> first( () )
    Traceback (most recent call last):
    ...
    StopIteration
    """

    return next(x for x in iterable if condition(x))



from datetime import datetime
import ipaddress
import logging
import os

import git
import gitdb
import jinja2
from pottery import Redlock

from osism.utils import first
from osism import utils


def vlans_as_string(untagged_vlan, tagged_vlans):
    vlans = [str(x.vid) for x in tagged_vlans]
    if untagged_vlan:
        vlans = vlans + [str(untagged_vlan.vid)]
    return ",".join(sorted(vlans))


def for_device(name, template=None):
    device = utils.nb.dcim.devices.get(name=name)

    if "device_type" not in device.custom_fields or device.custom_fields["device_type"] != "switch":
        return

    if "Managed by OSISM" not in [str(x) for x in device.tags]:
        return

    logging.info(f"Generate configuration for device {device.name}")

    if not template:
        template = f"{device.name}.cfg.j2"

        if not os.path.isfile(f"/netbox/templates/{template}"):
            template = f"{device.device_type.manufacturer.name}.cfg.j2"

        if not os.path.isfile(f"/netbox/templates/{template}"):
            template = "default.cfg.j2"

    vlans = utils.nb.ipam.vlans.filter(available_on_device=device.id)
    interfaces = utils.nb.dcim.interfaces.filter(device=device)

    interfaces_ethernet = []
    interfaces_port_channels = []
    interfaces_virtual = []

    for interface in interfaces:
        if str(interface.type) == "Link Aggregation Group (LAG)":
            interfaces_port_channels.append(interface)
        elif str(interface.type) == "Virtual":
            interfaces_virtual.append(interface)
        else:
            interfaces_ethernet.append(interface)

    # Port-Channel10 + Vlan4094 are always used as MLAG
    try:
        mlag = utils.nb.dcim.interfaces.get(device=device, name="Port-Channel10")
        mlag_vlan = utils.nb.dcim.interfaces.get(device=device, name="Vlan4094")
        mlag_address = utils.nb.ipam.ip_addresses.get(device=device, interface="Vlan4094")
    except:  # noqa
        mlag = None
        mlag_vlan = None
        mlag_address = None

    # NOTE: only work with /30
    try:
        x = list(ipaddress.ip_interface(mlag_address.address).network.hosts())
        x.remove(ipaddress.ip_interface(mlag_address.address).ip)
        mlag_peer_address = x[0]
    except:  # noqa
        mlag_peer_address = None

    try:
        mlag_domain_id = device.name.split("-")[1]
    except:  # noqa
        mlag_domain_id = None

    repo = git.Repo.init(path="/state")
    repo.config_writer().set_value("user", "name", "Netbox Generator").release()
    repo.config_writer().set_value("user", "email", "netbox-generator@reconciler.local").release()

    data = {
        "hostname": device.name,
        "interfaces_ethernet": interfaces_ethernet,
        "interfaces_virtual": interfaces_virtual,
        "interfaces_port_channels": interfaces_port_channels,
        "vlans": vlans,
        "device": name,
        "nb": utils.nb,
        "first": first,
        "vlans_as_string": vlans_as_string,
        "mlag": mlag,
        "mlag_vlan": mlag_vlan,
        "mlag_peer_address": mlag_peer_address,
        "mlag_domain_id": mlag_domain_id,
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "device_type": device.device_type.model,
        "device_manufacturer": device.device_type.manufacturer.name
    }

    loader = jinja2.FileSystemLoader(searchpath="/netbox/templates/")
    environment = jinja2.Environment(loader=loader)
    template = environment.get_template(template)
    result = template.render(data)

    logging.info(f"Writing generated configuration to /state/{device.name}.cfg.j2")
    with open(f"/state/{device.name}.cfg.j2", "w+") as fp:
        fp.write(os.linesep.join([s for s in result.splitlines() if s]))

    # Allow only one change per time
    lock = Redlock(key="lock_repository", masters={utils.redis})
    lock.acquire()

    repo.git.add(f"/state/{device.name}.cfg.j2")

    try:
        if len(repo.index.diff("HEAD")) > 0:
            logging.info(f"Committing changes in /state/{device.name}.cfg.j2")
            repo.git.commit(message=f"Update {device.name}")

    # Ref 'HEAD' did not resolve to an object
    except gitdb.exc.BadName:
        logging.info("Initial commit")
        repo.git.commit(message="Initial commit")

    lock.release()



import logging
import subprocess

from cliff.command import Command

from osism.tasks import reconciler


class Run(Command):

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        # NOTE: use python interface in the future, works for the moment
        p = subprocess.Popen("celery -A osism.tasks.reconciler worker -n reconciler --loglevel=INFO -Q reconciler", shell=True)
        p.wait()


class Sync(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Sync, self).get_parser(prog_name)
        parser.add_argument('--no-wait', default=False, help='Do not wait until the sync has been completed', action='store_true')
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        task = reconciler.run.delay()
        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)



import argparse
import logging

from cliff.command import Command
from redis import Redis

from osism.tasks import conductor, netbox, reconciler, ansible, openstack


redis = Redis(host="redis", port="6379")


class Run(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument('arguments', nargs=argparse.REMAINDER, help='Other arguments for Ansible')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the role has been applied', action='store_true')
        return parser

    def take_action(self, parsed_args):
        pass


class Bifrost(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Bifrost, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        ansible.run.apply_async(("manager", "bifrost-command", "baremetal node list -f json"), link=netbox.synchronize_device_state.s())


class Ironic(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Ironic, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        # Get Ironic parameters from the conductor
        task = conductor.get_ironic_parameters.delay()
        task.wait(timeout=None, interval=0.5)
        ironic_parameters = task.get()

        # Add all unregistered systems from the Netbox in Ironic
        netbox.get_devices_not_yet_registered_in_ironic.apply_async((), link=openstack.baremetal_create_nodes.s(ironic_parameters))

        # Synchronize the current status in Ironic with the Netbox
        # openstack.baremetal_node_list.apply_async((), link=netbox.synchronize_device_state.s())

        # Remove systems from Ironic that are no longer present in the Netbox


class Sync(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Sync, self).get_parser(prog_name)
        parser.add_argument('--no-wait', default=False, help='Do not wait until the sync has been completed', action='store_true')
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        task = reconciler.sync_inventory_with_netbox.delay()
        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Init(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Init, self).get_parser(prog_name)
        parser.add_argument('arguments', nargs=argparse.REMAINDER, help='Other arguments for Ansible')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the role has been applied', action='store_true')
        return parser

    def take_action(self, parsed_args):
        arguments = parsed_args.arguments
        wait = not parsed_args.no_wait

        task = ansible.run.delay("netbox-local", "init", arguments)

        if wait:
            self.log.info("Task is running. Wait. No more output. Check ARA for logs.")
            task.wait(timeout=None, interval=0.5)


class Import(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Import, self).get_parser(prog_name)
        parser.add_argument('--vendors', help='Vendors from which all available device types are to be imported', required=False)
        parser.add_argument(
            '--no-library',
            default=False,
            help='Do not import device types from the device type library, use the config repository',
            action='store_true'
        )
        parser.add_argument('--no-wait', default=False, help='Do not wait until the role has been applied', action='store_true')
        return parser

    def take_action(self, parsed_args):
        vendors = parsed_args.vendors
        wait = not parsed_args.no_wait
        library = not parsed_args.no_library

        task = netbox.import_device_types.delay(vendors, library)

        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Manage(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Manage, self).get_parser(prog_name)
        parser.add_argument('--type', type=str, help='Type of the resource to manage', required=False, default='rack')
        parser.add_argument('name', nargs=1, type=str, help='Name of the resource to manage')
        parser.add_argument('arguments', nargs=argparse.REMAINDER, help='Other arguments for Ansible')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the changes have been made', action='store_true')
        return parser

    def take_action(self, parsed_args):
        arguments = parsed_args.arguments
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait
        type_of_resource = parsed_args.type

        task = ansible.run.delay("netbox-local", f"{type_of_resource}-{name}", arguments)

        if wait:
            self.log.info("Task is running. Wait. No more output. Check ARA for logs.")
            task.wait(timeout=None, interval=0.5)


class Connect(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Connect, self).get_parser(prog_name)
        parser.add_argument('name', nargs='?', type=str, help='Name of the resource (a collection or a device) to connect')
        parser.add_argument('--collection', type=str, help='Name of the collection to connect', required=False),
        parser.add_argument('--device', type=str, help='Name of the device to connect', required=False)
        parser.add_argument('--enforce', default=False, help='Ignore the current transition of a device', action='store_true')
        parser.add_argument('--state', type=str, help='State to use', default=None, required=False)
        parser.add_argument('--type', type=str, default='collection', help='Type of the resource to connection (when not using --collection or --device)', required=False)
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name
        collection = parsed_args.device
        device = parsed_args.device
        state = parsed_args.state
        type_of_resource = parsed_args.type
        enforce = parsed_args.enforce

        task = None

        if name:
            if type_of_resource == "collection":
                task = netbox.data.delay(name, "", state)
            elif type_of_resource == "device":
                task = netbox.data.delay("", name, state)
        else:
            task = netbox.data.delay(collection, device, state)
            name = f"{collection}-{device}"

        task.wait(timeout=None, interval=0.5)
        data = task.get()

        self.log.info("Tasks are running in background. No more output. Check Flower for logs.")
        for device in data:
            netbox.connect.delay(device, state, data, enforce)


class Disable(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Disable, self).get_parser(prog_name)
        parser.add_argument('name', nargs=1, type=str, help='Name of the device to check for unused interfaces')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the changes have been made', action='store_true')
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait

        task = netbox.disable.delay(name)

        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Generate(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Generate, self).get_parser(prog_name)
        parser.add_argument('name', nargs=1, type=str, help='Name of the device for which the configuration is to be generated.')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the changes have been made', action='store_true')
        parser.add_argument('--template', type=str, help='Name of the template to use', required=False)
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        template = parsed_args.template
        wait = not parsed_args.no_wait

        task = netbox.generate.delay(name, template)

        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Deploy(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Deploy, self).get_parser(prog_name)
        parser.add_argument('name', nargs=1, type=str, help='Name of the device for which the configuration is to be deployed')
        parser.add_argument('arguments', nargs=argparse.REMAINDER, help='Other arguments for Ansible')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the changes have been made', action='store_true')
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        # arguments = parsed_args.arguments
        wait = not parsed_args.no_wait

        task = netbox.deploy.delay(name)

        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Check(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Check, self).get_parser(prog_name)
        parser.add_argument('name', nargs=1, type=str, help='Name of the device for which the configuration is to be checked')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the changes have been made', action='store_true')
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait

        task = netbox.check.delay(name)

        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Diff(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Diff, self).get_parser(prog_name)
        parser.add_argument('name', nargs=1, type=str, help='Name of the device for which the configuration is to be diffed')
        parser.add_argument('--no-wait', default=False, help='Do not wait until the changes have been made', action='store_true')
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait

        task = netbox.diff.delay(name)

        if wait:
            self.log.info("Task is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Ping(Command):

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super(Ping, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        task = netbox.ping.delay()
        task.wait(timeout=None, interval=0.5)
        result = task.get()
        print(result)
